__version__ = "0.3"
__name__ = "python-airtable"
__description__ = "Python module for accessing Airtable (largely based on the original airtable-python-wrapper)."
__url__ = "https://github.com/fivehealth/python-airtable"
__author__ = "5 Health Inc"
__authoremail__ = "engineers@botmd.io"
__license__ = "The MIT License (MIT)"
__copyright__ = "Copyright 2020 5 Health Inc"
